#!/usr/bin/node

"use strict"

const   BigNumber   =   require('bignumber.js');
const   Libbfa      =   require( process.env.BFAHOME + '/bin/libbfa.js');
const   rl          =   require('readline').createInterface(
        { input: process.stdin, output: process.stdout }
    );
var     web3;
var     Distillery;
var     bfa;
var     notation    =   { "potency": 15 };

function    init()
{
    var     table   =   [
         6, 'Mwei',
         9, 'Gwei',
        12, 'micro',
        15, 'finney',
        18, 'ether',
        21, 'kether',
        24, 'grand',
        27, 'mether',
        30, 'gether',
        33, 'tether'
    ];
    for (var i = 0; table[i]; i+=2)
        if ( table[i] == notation.potency )
            notation.name   =   table[i+1]
    if ( undefined == notation.name )
        notation    = { 'potency': 15, 'name': 'finney' };
    notation.num    = BigNumber( 10 ).pow( notation.potency );
    bfa             =   new Libbfa();
    web3            =   bfa.newweb3();
    Distillery      =   bfa.contract( web3, 'Distillery' );
    requestDistBalance()
    .then( requestBalances )
}

function    palletSort(a,b)
{
    if ( b == undefined )
    {
        if ( a == undefined )
            return 0;
        else
            return -1;
    }
    if ( a == undefined )
        return 1;
    var strA    =   a.addr.toLowerCase();
    var strB    =   b.addr.toLowerCase();
    if ( strA < strB )
        return -1;
    if ( strA > strB )
        return 1;
    return 0;
}

async function    requestDistBalance()
{
    return web3.eth.getBalance(Distillery.contractaddress)
    .then( (bal) => { return Distillery.contractbalance = new BigNumber( bal ) } )
}

async function    requestBalances()
{
    var     count   =   await Distillery.methods.numberOfBeneficiaries().call();
    var     pallet  =   new Array;
    var     i;
    // Fetch addresses from the list in the contract.
    for ( i=0; i<count; i++ )
    {
        var     addressandsetting    =   await Distillery.methods.atPosition(i).call();
        var     addr    =   addressandsetting[0];
        var     setting =   addressandsetting[1];
        var     bal     =   await web3.eth.getBalance( addr );
        pallet.push( { "addr": addr, "setting": new BigNumber(setting), "balance": new BigNumber(bal) } );
    }
    displayBalances(pallet);
}

function    editAccount( entry, pallet )
{
    if ( entry == undefined )
        return;
    var     acct;
    var     value;
    // it is an existing account address?
    if ( bfa.isAddr(entry) )
    {
        var     i       =   0;
        var     n       =   pallet.length;
        while ( i < n )
        {
            if (String(pallet[i].addr).toLowerCase() == String(entry).toLowerCase() )
                entry   =   i;
            i++;
        }
    }
    // it is a new account address?
    if ( bfa.isAddr(entry) )
    {
        acct            =   entry;
        value           =   0;
    }
    else
    if  ( bfa.isNumeric(entry) && entry < pallet.length )
    {
        acct            =   pallet[entry].addr;
        value           =   pallet[entry].setting;
    }
    else
    if ( entry == "x" )
    {
        var     rcpt;
        // trigger distribution
        Distillery.methods.distribute().send( {"from": bfa.account, "gas": 4000000, gasPrice: 1000000000 } )
        .on( 'error', (x) => { bfa.fatal( "Distribute returned errors: " + x ) } )
        .on( 'confirmation', (n,x) =>
            {
                if ( undefined == rcpt )
                {
                    rcpt    =   x;
                    console.log(
                        "Distribute returned succesfully in block# "
                        + rcpt.blockNumber
                        + " using "
                        + rcpt.gasUsed
                        + " gas."
                    );
                    var beforeBal = Distillery.contractbalance;
                    requestDistBalance()
                    .then(
                        function()
                        {
                            console.log(
                                "Distributed "
                                + beforeBal.minus( Distillery.contractbalance ).div( notation.num ).toFixed()
                                + " "
                                + notation.name
                                + "."
                            );
                            requestBalances();
                        }
                    )
                }
            }
        );
        return;
    }
    else
    if ( entry == "" )
    {
        // Do nothing, basically just update the display
        return;
    }
    else
    {
        bfa.fatal("I don't know what to do with \""+entry+"\"." );
    }
    rl.question(
        "Adjust the "
        + notation.name
        + " fill value of "
        + acct
        + " (setting to 0 is the same as deleting)\n"
        + "Amount?: ",
        (answer) => {
            if ( bfa.isNumeric(answer) )
            {
                var     bi  =   BigNumber( answer ).multipliedBy( notation.num );
                console.log("Sending update to the SC...");
                var     rcpt;
                Distillery.methods.setEtherAllowance(acct,web3.utils.toHex(bi))
                .send( {"from": bfa.account, gasPrice: 1000000000, gas: 100000 } )
                .on( 'error', (err) => {
                        bfa.fatal(
                            "\nMaybe you are not authorized:\n"
                            +err
                            +"\n\n\nI think you should leave now.\n"
                        );
                    }
                )
                .on( 'confirmation', (n,x) => {
                        if ( undefined == rcpt )
                        {
                            rcpt    =   x;
                            console.log("Update accepted.")
                            requestBalances();
                        }
                    }
                );
            }
            else
                bfa.fatal( "I have no idea what to do with \""+answer+"\"." );
            rl.close;
        }
    );
}

function    displayBalances( pallet )
{
    var     n       =   pallet.length;
    var     i;
    pallet.sort(palletSort);
    console.log(
        "The contract's account ("
        + Distillery.contractaddress
        + ") has "
        + Distillery.contractbalance.div( notation.num ).toFixed()
        + " "
        + notation.name
        + ".\n"
    );
    var     longest         =   1;
    for ( i=0; i<n; i++ )
    {
        var len =   pallet[i].setting.div(notation.num).toFixed(notation.potency).length;
        if ( len > longest )
            longest = len;
    }
    for ( i=0; i<n; i++ )
    {
        var     entry   =   pallet[i];
        if ( entry == undefined )
            console.log( i+": <undef>" );
        else
        {
            var     numstr  =   pallet[i].setting.div(notation.num).toFixed(notation.potency);
            while ( numstr.length < longest )
                numstr      =   " "+numstr;
            console.log(
                i
                + ": "
                + pallet[i].addr
                + " fills to "
                + numstr
                + " "
                + notation.name
                + " (has "
                + pallet[i].balance.div(notation.num).toFixed(notation.potency)
                + ")."
            );
        }
    }
    console.log("\n[ Q=quit x=distribute ]");
    rl.question("Which account to edit (enter index number of full account number)?: ",
        (answer) => {
            if ( answer != undefined )
            {
                if ( String(answer).toUpperCase() == 'Q' )
                    process.exit( 0 );
                else
                    editAccount(answer, pallet);
            }
            rl.close;
        }
    )
}

init();
